# Generated by Django 2.2.10 on 2021-02-02 09:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("storage", "0008_accesslog_cause"),
    ]

    operations = [
        migrations.AddField(
            model_name="referencedpath",
            name="chunk_size",
            field=models.IntegerField(default=8388608),
        ),
    ]
